/******************************************************************************
 CIconPane.c

		A pane that can draw an ICON or CICN resource, and can behave as 
		a button.
				
	SUPERCLASS = CPane
	
	Copyright  1991 Symantec Corporation. All rights reserved.
	

 ******************************************************************************/

#include "CIconPane.h"
#include "OSChecks.h"
#include "Commands.h"
#include "Global.h"
#include "OSChecks.h"


	/* This bit map and buffer are used for drawing the icon in		*/
	/* either normal or hilite mode									*/
	
BitMap	CIconPane::cMaskBitMap;
char 	CIconPane::cIconMaskBits[kIconBytes];

/*
 *	TCL 1.1.2 DLP 12/10/91
 *	When available, we now use the System 7 Icon Utilities for drawing cicns
 *  See TN #306 for more information
 */
Boolean CIconPane::cHaveIconDispatch = FALSE;

#define tcl_IconDispatch	0xABC9
pascal OSErr tcl_PlotCIconHandle( Rect*, short, short, CIconHandle) = { 0x303C, 0x61F, tcl_IconDispatch};

/******************************************************************************
 IIconPane
 
 	Initialize an IconPane. The icon is taken from an 'ICON' resource
 	with the specified ID. If the fPreferColor flag is TRUE,
 	Color QuickDraw is available, and a 'cicn' resource with the same
 	ID exists, then the color icon will be used instead.
******************************************************************************/

void CIconPane::IIconPane( CView *anEnclosure, CBureaucrat *aSupervisor,
					short aHEncl, short aVEncl,
					SizingOption aHSizing, SizingOption aVSizing,
					short iconID, Boolean fPreferColor)
{	
	CPane::IPane( anEnclosure, aSupervisor, kIconPixels, kIconPixels, aHEncl, aVEncl,
				aHSizing, aVSizing);
	
	this->iconID = iconID;
	cicnH = NULL;
	clickCmd = cmdNull;

	IIconPaneX( fPreferColor);
	
}	/* CIconPane::IIcon */

/******************************************************************************
 IViewTemp
 
 	Initialize from a template resource.
******************************************************************************/

void CIconPane::IViewTemp(CView *anEnclosure, CBureaucrat *aSupervisor,
						Ptr viewData)
{
	tIconPaneTempP template = (tIconPaneTempP) viewData;
	
	inherited::IViewTemp( anEnclosure, aSupervisor, viewData);
	iconID = template->iconID;
	clickCmd = template->clickCmd;
	cicnH = NULL;
	
	IIconPaneX( template->fPreferColor);

}	/* CIconPane::IViewTemp */

/******************************************************************************
 IIconPaneX
 
 	Perform common initialization.
******************************************************************************/

void CIconPane::IIconPaneX( Boolean fPreferColor)
{
	Handle	iconH;
		
	/* If color is desired and Color QD is available, then look for	*/
	/* a matching 'cicn' resource.									*/
	
	if (fPreferColor && gSystem.hasColorQD)
	{
		cicnH = GetCIcon( iconID);
		if (cicnH) MoveHHi( (Handle) cicnH);
	}

	iconH = GetIcon( iconID);
	FailNILRes( iconH);

	BlockMove( *iconH, icon, kIconBytes);
	
		/* set up the bitmap	*/
	
	cMaskBitMap.baseAddr = cIconMaskBits; /* point to the mask bits	*/
	cMaskBitMap.rowBytes = kIconRowBytes;
	SetRect( &cMaskBitMap.bounds, 0, 0, kIconPixels, kIconPixels);
	
		/* TCL 1.1.2 DLP 12/10/91: check for Icon Utilities trap */
		
	cHaveIconDispatch = TrapAvailable( tcl_IconDispatch);
	
}	/* CIconPane::IIconPaneX */

/******************************************************************************
 DrawIcon
 
 	Draw the icon normally or hilited.
 	
 	TCL 1.1.2: DrawIcon has been substantially changed to use the
 		System 7 Icon Utilities for cicns, and to preserve the
 		foreground and background colors of the port.
******************************************************************************/

void CIconPane::DrawIcon( Boolean fHilite)
{
	Rect		theBounds = { 0, 0, kIconPixels, kIconPixels};
	short		copyMode = fHilite ? srcOr : srcBic;
	RGBColor	fore, back;
	
	if (gSystem.hasColorQD)	// preserve foreground and background colors
	{
		GetForeColor( &fore);
		GetBackColor( &back);
	}
	ForeColor( blackColor);	// set default colors for CopyBits
	BackColor( whiteColor);	
		
	if (cicnH)
	{
		if (cHaveIconDispatch)
		{
			// under System 7 or later, let the Icon Utilities do the
			// correct highlighting
			
			short iconTransform = fHilite? 0x4000 : 0;Debugger();
			FailOSErr(tcl_PlotCIconHandle( &theBounds, 0, iconTransform, cicnH));
		}
		else
		{	
			// To draw the cicn highlighted, we make a copy of the cicn's
			// pixmap color table and "dim" it by modifying the RGB values.	
			if (fHilite)
			{
				CTabHandle	ctab = (**cicnH).iconPMap.pmTable, dupCTable;
				short		i, numColors;
				ColorSpec	*cspec;

				dupCTable = ctab;
				
					// let this memory allocation come from the reserve if necessary.
					// we'll be giving it back almost immediately.
				HandToHand( &dupCTable);
				
				numColors = (**ctab).ctSize; // number of colors in cicn's color table
				cspec = (**ctab).ctTable;	// point to first ColorSpec in the color table
				for(i = 0; i < numColors; i++)
				{
					cspec->rgb.red /= 2;	// dim by halving the RGBs
					cspec->rgb.green /= 2;
					cspec->rgb.blue /= 2;
					++cspec;
				}
				PlotCIcon( &theBounds, cicnH); // plot it dimmed
				
					// put back the undimmed color table, then free the dimmed one.
				(**cicnH).iconPMap.pmTable = dupCTable;
				DisposCTable( ctab);
			}
			else	// no dimming, just plot it normally
			
				PlotCIcon( &theBounds, cicnH);
		}
	}
	else	// draw the b&w icon
	{
		BitMap  theImage;
		Boolean wasLocked = Lock( TRUE);
		theImage.baseAddr = icon;
		theImage.rowBytes = kIconRowBytes;
		
		theImage.bounds = theBounds;
				
		CalcMask( icon, cIconMaskBits, kIconRowBytes, kIconRowBytes, kIconPixels, 2);

		CopyBits( &cMaskBitMap, &thePort->portBits, &theBounds, &theBounds, copyMode, NULL);
		CopyBits( &theImage, &thePort->portBits, &theBounds, &theBounds, srcXor, NULL); 
	 
		Lock( wasLocked);	
	}
	if (gSystem.hasColorQD)	// restore the colors
	{
		RGBForeColor( &fore);
		RGBBackColor( &back);
	}

}	/* CIconPane::DrawIcon */

/******************************************************************************
 Draw
 
 	Draw the icon in response to an update.
******************************************************************************/

void  CIconPane::Draw( Rect *area)
{
	DrawIcon( FALSE);

}	/* CIconPane::Draw */


/******************************************************************************
 SetClickCmd
 
 	Set the clickCmd for an iconPane. This command will be sent to
 	itsSupervisor when the icon is clicked and the mouse is released
 	inside the icon.
******************************************************************************/

void CIconPane::SetClickCmd( long aCmd)
{
	clickCmd = aCmd;

}	/* CIconPane::SetClickCmd */

/******************************************************************************
 GetClickCmd
 
 	Return a CIconPane's clickCmd
******************************************************************************/
long CIconPane::GetClickCmd( void)
{
	return clickCmd;

}	/* CIconPane::GetClickCmd */

/******************************************************************************
 DoClick
 
 	Respond to a click. The wantsClicks instance variable must be
 	TRUE for this method ever to be called.
******************************************************************************/

void CIconPane::DoClick( Point hitPt, short modifierKeys, long when)
{
	if (Track())
	{
		itsSupervisor->DoCommand( GetClickCmd());
	}

}	/* CIconPane::DoClick */

/******************************************************************************
 Track
 
 	Tracks the mouse, returns TRUE if the mouse was in the icon when released.
******************************************************************************/

Boolean CIconPane::Track( void)
{
	Boolean	inBtn = TRUE;
	Point	where;
	Rect	qdFrame;
		
	DrawIcon( TRUE);
	LongToQDRect( &frame, &qdFrame);
	
	while (StillDown())
	{
		GetMouse( &where);
		if (PtInRect( where, &qdFrame))
		{
			if (!inBtn) DrawIcon( TRUE);
			inBtn = TRUE;
		}
		else
		{
			if (inBtn) DrawIcon( FALSE);
			inBtn = FALSE;
		}
	}
	if (inBtn) DrawIcon( FALSE);
	return inBtn;

}	/* CIconPane::Track */

/******************************************************************************
 Dispose
 
 	Free the color icon if it exists
******************************************************************************/

void  CIconPane::Dispose( void)
{
	if (cicnH)
	{
		DisposCIcon( cicnH);
		cicnH = NULL;
	}
	inherited::Dispose();

}	/* CIconPane::Dispose */
