/******************************************************************************
 CMenuDefProc.c

							The MenuDefProc Class
		
		Abstract class for a "Menu Definition Procedure". Not really a
		single procedure, but a class with methods corresponding to each
		of the actions the Toolbox can request of a menu.
	
	SUPERCLASS = CObject
	
	Copyright  1989 Symantec Corporation. All rights reserved.

 ******************************************************************************/

#include "Global.h"
#include "CMenuDefProc.h"
#include "GenericMDEF.h"
#include "OSChecks.h"	/* TCL 1.1.1 DLP 9/25/91 */

/******************************************************************************
 IMenuDefProc

		Initialize a MenuDefProc object
 ******************************************************************************/

void	CMenuDefProc::IMenuDefProc(
	short		MDEFid)
{
	GenericMDEFHand	theMDEF;
	
	theMDEF = (GenericMDEFHand) GetResource('MDEF', MDEFid);
	FailNILRes( theMDEF);
	(**theMDEF).defProc = (VoidFunc) GenericMDEF;
	(**theMDEF).itsMenuDefProc = this;
	FlushCache();	/* TCL 1.1.1 DLP 9/25/91 */
					/* After modifying code, we must flush the cache on 68040s */
}
	

/******************************************************************************
 DrawMenu

		Draw the contents of the menu. Subclasses must override this
		method.
 ******************************************************************************/

void	CMenuDefProc::DrawMenu(
	MenuHandle		macMenu,
	Rect			*menuRect)
{
}										/* Null Method						*/
	

/******************************************************************************
 ChooseItem

		Determine which item in the menu is hit. Subclasses which support
		selection of items from the menu must override this method.
 ******************************************************************************/

void	CMenuDefProc::ChooseItem(
	MenuHandle		macMenu,
	Rect			*menuRect,
	Point			hitPt,
	short			*whichItem)
{
	*whichItem = NOTHING;
}
	

/******************************************************************************
 SizeMenu

		Determine the height and width of the menu in pixels. Subclasses
		must override this method.
 ******************************************************************************/

void	CMenuDefProc::SizeMenu(
	MenuHandle		macMenu)
{
	(**macMenu).menuWidth = 0;
	(**macMenu).menuHeight = 0;
}
	

/******************************************************************************
 PlacePopup

		Determine the placement of a pop-up menu. Subclasses must override
		this method if they wish to support pop-up menus.
 ******************************************************************************/

void	CMenuDefProc::PlacePopup(
	MenuHandle		macMenu,
	Rect			*menuRect,
	Point			hitPt,
	short			*whichItem)
{
}										/* Null Method						*/
