/******************************************************************************
 TBUtilities.c
						Utilties for dealing with the Toolbox
								
		Exports:
			Boolean		IsSystemWindow(macWindow);
			Boolean		IsMyWindow(macWindow);
			void		BringBehind(macWindow, behindWindow);
			Boolean		IsDialogWindow(macWindow);
			void		PositionDialog(theType, theID);
			void		GetFontNumber(Str255 fName, short* itemNum);
			void		CopyPString(Str255 scrString, Str255 destString);
			void		ConcatPStrings(Str255 first, Str255 second);
			Boolean		KeyIsDown(short theKeyCode);
			void		DrawSICN(short SICNid, short index Point location);
			void		PinInRect( LongRect *theRect, LongPt *thePoint);
			Boolean		AbortInQueue( void);
			Boolean		IsCancelEvent( EventRecord *);

			
		Copyright  1989 Symantec Corporation. All rights reserved.
		
	TCL 1.1 CHANGES
	[
		- implemented AbortInQueue, IsCancelEvent functions
	]

 ******************************************************************************/

#include "Global.h"
#include "TBUtilities.h"
#include "CDesktop.h"
#include "Constants.h"
#include "LongCoordinates.h"

#include <Script.h>

extern CDesktop		*gDesktop;
/******************************************************************************
 IsSystemWindow

		Determine if the window belongs to a DA
 ******************************************************************************/

Boolean	IsSystemWindow(
	WindowPeek		macWindow)			/* The window in question			*/
{
		/* System windows (DAs) have a negative windowKind */
		
	return( macWindow && (macWindow->windowKind < 0) );
}


/******************************************************************************
 IsMyWindow

		Check whether a window is an application window
 ******************************************************************************/

Boolean	IsMyWindow(
	WindowPeek		macWindow)			/* The window in question			*/
{
		/* All application windows have a kind >= userKind */

	return ( macWindow && 
			 ( (macWindow->windowKind >= userKind)  ||
			   (macWindow->windowKind == dialogKind) ) );
}


/******************************************************************************
 BringBehind

		Move a window from far back to right behind another window
 ******************************************************************************/

void	BringBehind(
	register WindowPtr	macWindow,		/* The window to move				*/
	WindowPtr			behindWindow)	/* Window to put the window behind	*/
{
	extern RgnHandle	gUtilRgn;		/* Originally visible window parts	*/
										/*  AND newly exposed window parts	*/
	GrafPtr				savePort;		/* Current port						*/
	Point				corner;			/* Top left of visible region		*/
	
	GetPort(&savePort);					/* Save current port				*/
	SetPort(macWindow);					/* Use this window's port			*/
	
										/* Save portion of window which is	*/
										/* originally visible				*/
	CopyRgn(macWindow->visRgn, gUtilRgn);
										/* Adjust the window's plane		*/
	SendBehind(macWindow, behindWindow);

		/* We must draw the newly exposed portion of the window.  Find the */
		/* difference between the present structure region and what was    */
		/* originally visible.  Before doing this, we must convert the     */
		/* originally visible region to global coords.					   */
	
	corner = topLeft((**gUtilRgn).rgnBBox);
	LocalToGlobal(&corner);
	OffsetRgn(gUtilRgn, (corner.h - (**gUtilRgn).rgnBBox.left),
						(corner.v - (**gUtilRgn).rgnBBox.top));
					  
		/* Now we can difference the regions.  Save space by putting the */
		/* result back in theRgn.  Before calling DiffRgn, theRgn is the */
		/* originally visible region.  Afterwards, theRgn is the newly   */
		/* exposed region of the window.								 */
		
	DiffRgn(((WindowPeek)macWindow)->strucRgn, gUtilRgn, gUtilRgn);
	
	PaintOne((WindowPeek) macWindow, gUtilRgn);		/* Draw newly exposed region		*/
	
										/* Since window has moved forward,	*/
										/*   we must adjust the visible		*/
										/*   regions of this window and		*/
										/*   those behind it.				*/
	CalcVisBehind( (WindowPeek) macWindow, ((WindowPeek)macWindow)->strucRgn);

	SetPort(savePort);					/* Restore the original port		*/
	CView::ForceNextPrepare();
}


/******************************************************************************
 IsDialogWindow

		Determine if the window is a dialog box
 ******************************************************************************/

Boolean	IsDialogWindow(
	WindowPeek		macWindow)			/* The window in question			*/
{
		/* Dialog boxes have a windowKind of dialogKind */
		
	return( macWindow && (macWindow->windowKind == dialogKind) );
}


/******************************************************************************
 PositionDialog
 
		Center the bounding box of a dialog or alert in the upper third
		of the screen.  This is the preferred location according to the
		Human Interface Guidelines.
 ******************************************************************************/

void	PositionDialog(
	ResType		theType,				/* Resource type, DLOG or ALRT		*/
	short		theID)					/* Res ID of DLOG or ALRT			*/
{
	Handle				theTemplate;	/* Handle to resource template		*/
	register Rect		*theRect;		/* Bounding box of dialog			*/
	register short		left;			/* Left side of centered rect		*/
	register short		top;			/* Top side of centered rect		*/
	
		/* The first field of the resource template for DLOG's and ALRT's */
		/* is its bounding box.  Get a pointer to this rectangle.  This   */
		/* handle dereferencing is safe since the remaining statements in */
		/* this function do not move memory (assignment and simple math). */

	theTemplate = GetResource(theType, theID);
	FailNILRes( theTemplate);
	theRect = (Rect*) *theTemplate;
	
										/* Center horizontally on screen	*/
	left = (screenBits.bounds.right - (theRect->right - theRect->left)) / 2;

										/* Leave twice as much space below	*/
										/*   as above the rectangle			*/	
	top = (screenBits.bounds.bottom - (theRect->bottom - theRect->top)) / 3;
										/* Don't put rect under menu bar	*/
	top = Max(top, GetMBarHeight() + 1);

	theRect->right += left - theRect->left;
	theRect->left = left;
	theRect->bottom += top - theRect->top;
	theRect->top = top;
}


/******************************************************************************
 FindDlogPosition
 
 		Return the coordinates of the top left corner of a dialog or alert
 		which centers the box in the upper third of the main screen. This is
 		the preferred location according to the Human Interface Guidelines.
 ******************************************************************************/

void	FindDlogPosition(
	ResType		theType,				/* Resource type, DLOG or ALRT		*/
	short		theID,					/* Res ID of DLOG or ALRT			*/
	Point		*corner)				/* Top left corner when centered	*/
{
	Rect				theRect;		/* Bounding box of dialog			*/
	register short		left;			/* Left side of centered rect		*/
	register short		top;			/* Top side of centered rect		*/
	
		/* The first field of the resource template for DLOG's and ALRT's */
		/* is its bounding box.  Get a pointer to this rectangle.  This   */
		/* handle dereferencing is safe since the remaining statements in */
		/* this function do not move memory (assignment and simple math). */

	theRect = **((Rect**) GetResource(theType, theID));
	
										/* Center horizontally on screen	*/
	corner->h = (screenBits.bounds.right - (theRect.right - theRect.left)) / 2;

										/* Leave twice as much space		*/
										/*   as above the rectangle			*/	
	corner->v = (screenBits.bounds.bottom - (theRect.bottom - theRect.top)) / 3;
										/* Don't put rect under menu bar	*/
	corner->v = Max(corner->v, GetMBarHeight() + 1);
}


/******************************************************************************
 GetFontNumber
 
 	Find font number given the font name.  If not found, a negative
 	font number is returned.
 ******************************************************************************/

void	GetFontNumber(
	ConstStr255Param fontName,				/* Name of the font to find			*/
	short		*fontNum)				/* The font number					*/
{
	Str255		sysFontName;

	GetFNum(fontName, fontNum);			/* Find corresponding font number	*/
	
	if (*fontNum == systemFont) {

			/* fontNum is set to sysFont if fontName is not found.	*/
			/* We must check for the special case where the system	*/
			/* font is indeed being retrieved, by comparing the		*/
			/* name of the font with that of the system font.		*/
			
		GetFontName(systemFont, sysFontName);
		if (!EqualString(fontName, sysFontName, FALSE, FALSE))
										/* Names are NOT the same			*/
			*fontNum = -1;				/* Return a failure flag			*/
	}
}



/******************************************************************************
 CopyPString
 
		Copy a Pascal string
 ******************************************************************************/

void	CopyPString(
	ConstStr255Param		srcString,
	Str255		destString)
{
	BlockMove(srcString, destString, srcString[0] + 1L);
}


/******************************************************************************
 ConcatPStrings
 
		Concatenate two Pascal strings by attaching the second string on
		the end of the first string.
 ******************************************************************************/

void	ConcatPStrings(
	Str255		first,
	ConstStr255Param second)
{
    
    short charsToCopy;
    
    /* Truncate if concatenated string would be longer than 255 chars.  */
    
    charsToCopy = Min(second[0], 255 - first[0]);
    BlockMove(second + 1, first + first[0] + 1, (long) charsToCopy);
    first[0] += charsToCopy;
}


/******************************************************************************
 KeyIsDown
 
		Determine whether or not the specified key is being pressed. Keys
		are specified by hardware-specific key code (NOT the character).
 ******************************************************************************/

Boolean		KeyIsDown(
	short			theKeyCode)
{
	KeyMap			theKeys;
	
	GetKeys( theKeys);					/* Get state of each key			*/
										
		/* Ordering of bits in a KeyMap is truly bizarre. A KeyMap is a	*/
		/* 16-byte (128 bits) array where each bit specifies the start	*/
		/* of a key (0 = up, 1 = down). We isolate the bit for the		*/
		/* specified key code by first determining the byte position in	*/
		/* the KeyMap and then the bit position within that byte.		*/
		/* Key codes 0-7 are in the first byte (offset 0 from the		*/
		/* start), codes 8-15 are in the second, etc. The BitTst() trap	*/
		/* counts bits starting from the high-order bit of the byte.	*/
		/* For example, for key code 58 (the option key), we look at	*/
		/* the 8th byte (7 offset from the first byte) and the 5th bit	*/
		/* within that byte.											*/
		
	return( BitTst( ((char*) &theKeys) + theKeyCode / 8,
					(long) 7 - (theKeyCode % 8) ) );
}


/******************************************************************************
 DrawSICN
 
		Draw a SICN, a resource type which defines a 16 by 16 bit image,
		at the given location. SICN is specified by a resource ID number
		and a one-based index into the SICN list. The location specifies
		the top left corner of the SICN in the local coordinates of the
		current port.
 ******************************************************************************/

void	DrawSICN(
	short		SICNid,
	short		index,
	Point		location)
{
	Handle		theSICN;
	BitMap		theImage;
	Rect		theBounds;
	
	theSICN = GetResource('SICN', SICNid);
	FailNILRes( theSICN);
	
	HLock(theSICN);
										/* A SICN resource is really a list	*/
										/*   of small icons, each of which	*/
										/*   has 32 bytes of data			*/
										/*   16 x 16 = 256 bits = 32 bytes	*/
										
										/* Create a bitmap so we can copy	*/
										/*   the SICN image on the screen	*/
										
										/* Index into handle to get a ptr	*/
										/*   to the SICN bit image			*/
	theImage.baseAddr = *theSICN + (index - 1) * 32;
	theImage.rowBytes = 2;				/* 16 bits is 2 bytes				*/
	
										/* Place bitmap at desired locaton	*/
	SetRect(&theBounds, location.h, location.v,
						location.h + 16, location.v + 16);
	theImage.bounds = theBounds;
	
										/* Copy image on to the screen		*/
	CopyBits(&theImage, &thePort->portBits,
			 &theBounds, &theBounds, srcCopy, NULL);
			 
	HUnlock(theSICN);
}


/******************************************************************************
 PinInRect
 
		Pin a point inside a rectangle. Similar to the PinRect Toolbox
		trap except that the point is changed in place and one (1) is not
		subtracted at the right and bottom edges.
 ******************************************************************************/

void	PinInRect(
	const LongRect *theRect,
	LongPt	 *thePoint)
{
	thePoint->h = Max(theRect->left, thePoint->h);
	thePoint->h = Min(theRect->right, thePoint->h);
	thePoint->v = Max(theRect->top, thePoint->v);
	thePoint->v = Min(theRect->bottom, thePoint->v);
}


/******************************************************************************
 AbortInQueue			TCL 1.1.1 DLP 9/18/91
 
 	Walk the event queue, return TRUE if a Command-'.' is there. The
 	event is removed from the queue.
 	
 ******************************************************************************/

Boolean	AbortInQueue( void)
{
	EvQElPtr	qEvt = (EvQElPtr) GetEvQHdr()->qHead;
	Boolean		foundAbort = FALSE;

	while (qEvt)
	{
			// Starting at the evtQWhat field, and EvQEl is identical
			// to an EventRecord
			
		if ( IsCancelEvent( (EventRecord*) &qEvt->evtQWhat))
		{
			// Flush this and all other key-downs from the queue
			FlushEvents( keyDownMask, 0);
			foundAbort = TRUE;
			break;
		}
		qEvt = (EvQElPtr) qEvt->qLink;
	}
	return foundAbort;
}

/******************************************************************************
 IsCancelEvent
 
 	Returns TRUE if event is a command-period keydown event. Uses 
 	the technique in TN #263 to handle international keyboards.
 	
 ******************************************************************************/


#define kMaskModifiers  0xFE00     // we need the modifiers without the
                                   // command key for KeyTrans
                        
#define kUpKeyMask      0x0080
#define kMaskASCII1     0x00FF0000 // get the key out of the ASCII1 byte
#define kMaskASCII2     0x000000FF // get the key out of the ASCII2 byte

Boolean IsCancelEvent( EventRecord *theEvent )
{
	Boolean  isCancel = FALSE;
	short    keyCode;
	long     virtualKey, keyInfo, lowChar, highChar, state, keyCId;
	Handle   hKCHR = NULL;
	Ptr 	 KCHRPtr;

	if ((theEvent->what == keyDown) || (theEvent->what == autoKey))
	{
  
		// see if the command key is down.  If it is, find out the ASCII
  		// equivalent for the accompanying key.

		if (theEvent->modifiers & cmdKey) 
		{
				// get virtual key and keycode for KeyTrans
				
			virtualKey = (theEvent->message & keyCodeMask) >> 8;
	    
		    // and out the command key and or in the virtualKey
		    
		    keyCode = (theEvent->modifiers & kMaskModifiers)  |  virtualKey;
		   
		   	// make it look like a keyup event, to prevent dead key processing
		   	
		   	keyCode |= kUpKeyMask;
		    
		    KCHRPtr = (Ptr) GetEnvirons( smKCHRCache);
		
		    if (!KCHRPtr) 
		    {
				keyCId = GetScript( GetEnvirons(smKeyScript), smScriptKeys );
				hKCHR = GetResource( 'KCHR', keyCId );
				KCHRPtr = *hKCHR;
			}
			
			if (KCHRPtr) 
			{
				state = 0;
				keyInfo = KeyTrans( KCHRPtr, keyCode, &state);
				if (hKCHR)
					ReleaseResource(hKCHR);
			}
			else
				keyInfo = theEvent->message;
		
		    lowChar = keyInfo &  kMaskASCII2;
		    highChar = (keyInfo & kMaskASCII1) >> 16;
		    if (lowChar == '.' || highChar == '.')
				isCancel = true;
		}
	}
	return isCancel;
}
